"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import asdict, dataclass, field
import json
from tempfile import NamedTemporaryFile
from typing import TextIO
from proton.vpn.daemon.split_tunneling.apps.process_matcher import Process


@dataclass
class ProcessTreeNode:
    """Used to generate the process tree"""
    process: Process
    children: list[ProcessTreeNode] = field(default_factory=list)


class ProcessMap:
    """Utility class to dump tracked processes to disk for debugging purposes."""

    def __init__(self, processes: dict[int, Process]):
        self.processes = processes

    def dump(self) -> str:
        """
        Dumps the dict containing tracked processes to a temporary file.
        :returns: the path to the generated dump file.
        """
        with NamedTemporaryFile(prefix="process-dump-", mode="w", delete=False) as file:
            json.dump(
                [asdict(process) for process in self.processes.values()],
                file,
                indent=4,
                default=list  # to write python sets as json lists - json doesn't have sets
            )
            return file.name

    @staticmethod
    def load(path: str) -> ProcessMap:
        """
        Loads the process dump from the given path.
        :returns: the processes indexed by pid.
        """
        with open(path, encoding="utf8") as file:
            return ProcessMap(processes={
                process["pid"]: Process(
                    uid=process["uid"],
                    pid=process["pid"],
                    ppid=process["ppid"],
                    exe=process["exe"],
                    matched_config_paths=set(process["matched_config_paths"]),
                    running=process["running"]
                )
                for process in json.load(file)
            })

    def _get_root_processes(self) -> list[Process]:
        """
        :returns: the list of root processes.
        """
        return [
            process for process in self.processes.values()
            if process.ppid not in self.processes
        ]

    def _get_process_children(self, parent: Process) -> list[Process]:
        """
        :param parent: process to get the children from.
        :returns: the direct children of the parent process.
        """
        return [
            process for process in self.processes.values()
            if process.ppid == parent.pid
        ]

    def _get_process_tree(self, process: Process) -> ProcessTreeNode:
        """
        :param process: top process to generate its tree.
        :returns: the process tree of the specified top process.
        """
        return ProcessTreeNode(
            process=process,
            children=[
                self._get_process_tree(process=child)
                for child in self._get_process_children(process)
            ]
        )

    def get_process_trees(self) -> list[ProcessTreeNode]:
        """
        :returns: all process trees.
        """
        return [
            self._get_process_tree(process) for process in self._get_root_processes()
        ]

    def dump_process_tree(self, node: ProcessTreeNode, output_file: TextIO, nested_level: int):
        """
        Dumps a process tree to a file.
        :param node: the top tree node.
        :output_file: the file to dump the tree to.
        :nested_level: amount of indentation to dump the node with.
        """
        tabs = "\t" * nested_level
        output_file.write(f"{tabs}{node.process}\n")
        nested_level += 1
        for child in node.children:
            self.dump_process_tree(child, output_file, nested_level)

    def dump_process_trees(self) -> str:
        """
        Dumps all trees of processes to a temp file.
        :returns: the dump file path.
        """
        with NamedTemporaryFile(prefix="process-tree-", mode="w", delete=False) as file:
            process_trees = self.get_process_trees()
            for tree in process_trees:
                self.dump_process_tree(node=tree, output_file=file, nested_level=0)

            return file.name



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn import logging

logger = logging.getLogger(__name__)


def asyncio_exception_handler(_, context: dict):
    """Handles all asyncio exceptions.
    Asyncio exception handler as specified in:
    https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.set_exception_handler

    Args:
        _: The loop object.
        context (dict): Contains
    """
    exception = context.get("exception")

    if exception:
        exc_type = type(exception)
        exc_value = exception
        exc_traceback = context.get("traceback")

        logger.error(
            "Unhandled error.",
            exc_info=(exc_type, exc_value, exc_traceback)
        )



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Dict
from dbus_fast import Variant
from proton.vpn.core.settings import SplitTunnelingConfig, SplitTunnelingMode


def to_dbus_dict(data: SplitTunnelingConfig) -> Dict[str, object]:
    """Convert object to dbus compatible dict.

    When passing through dbus, we can't pass a python object
    thus we need to convert it to a dbus friendly dict.

    Returns:
        dict: keys are strings and values are `Variant`
    """
    return {
        "mode": Variant("s", data.mode.value),
        "app_paths": Variant("as", data.app_paths),
        "ip_ranges": Variant("as", data.ip_ranges),
    }


def from_dbus_dict(data: Dict[str, object]) -> SplitTunnelingConfig:
    """Generates `Config` from dbus dict.

    Args:
        data: keys are strings and values are `Variant`

    Returns:
        SplitTunnelingConfig: new config object
    """
    return SplitTunnelingConfig(
        mode=SplitTunnelingMode(data["mode"].value),
        app_paths=data["app_paths"].value,
        ip_ranges=data["ip_ranges"].value
    )



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio

from typing import Awaitable

from proton.vpn.daemon.split_tunneling.apps.process_monitor import \
    Process, ProcessMonitor, build_process_monitor_cli_parser
from proton.vpn.daemon.split_tunneling.apps.socket_monitor import SocketMonitor

from proton.vpn import logging
from proton.vpn.core.settings import SplitTunnelingConfig, SplitTunnelingMode

logger = logging.getLogger(__name__)


class AppBasedSplitTunnelingService:
    """Service to split-tunnel applications."""

    def __init__(self):
        self._process_monitor = ProcessMonitor(self._on_process_event)
        self._socket_monitor = SocketMonitor()

    def _on_process_event(self, process: Process, mode: SplitTunnelingMode):
        if process.matched_config_paths:
            logger.info("Process match: %s", process)

        if process.running and (
            (mode == SplitTunnelingMode.EXCLUDE and process.matched_config_paths) or
            (mode == SplitTunnelingMode.INCLUDE and not process.matched_config_paths)
        ):
            self._socket_monitor.exclude_process_from_vpn(process.pid)
        else:
            self._socket_monitor.stop_tracking_process(process.pid)

    def log_status(self):
        """Logs the service status."""
        logger.info("==============App-based ST service status============")
        self._process_monitor.log_status()
        self._socket_monitor.log_status()
        logger.info("=====================================================")

    def start(self, config_by_uid: dict[int, SplitTunnelingConfig]) -> Awaitable[None]:
        """
        Starts the service in the background. This method is non-blocking.
        :param config_by_uid: split tunneling configuration indexed by unix user ID.
        :returns: an awaitable to be able to await until the service is stopped.
        """
        self._socket_monitor.start()
        return self._process_monitor.start(config_by_uid=config_by_uid)

    async def stop(self):
        """Stops the service."""
        try:
            await self._process_monitor.stop()
        finally:
            self._socket_monitor.stop()


async def main():
    """Test script"""

    parser = build_process_monitor_cli_parser(
        name="App-based Split Tunneling"
    )
    args = parser.parse_args()

    logging.config(filename="app-based-st-service")

    service = AppBasedSplitTunnelingService()
    try:
        await service.start(config_by_uid={
            args.uid: SplitTunnelingConfig(
                mode=SplitTunnelingMode(value=args.mode),
                app_paths=args.path
            )
        })
    except asyncio.CancelledError:
        pass
    finally:
        try:
            await service.stop()
        except asyncio.CancelledError:
            pass


if __name__ == "__main__":
    asyncio.run(main())
